@@ -35,6 +35,26 @@ def concat(arrs, dim=0):
     
     return Array(np.concatenate(arrs, axis=dim))
 
+def stack(arrs, dim=0):
+    if len(arrs) == 0:
+        raise ValueError(f"arrs must be at least length 1, but {len(arrs)=}.")
+
+    is_tensor = [Array(arr).is_tensor for arr in arrs]
+    if not lib.np.isconst(is_tensor):
+        raise TypeError(f"Underlying data of arrs must either all be torch.Tensor or all be np.ndarray, but {is_tensor=}.")
+
+    is_masked_array = [Array(arr).is_masked_array for arr in arrs]
+    is_tensor = is_tensor[0]
+    arrs = [arr.data for arr in arrs]
+    
+    if is_tensor:
+        return Array(torch.stack(arrs, dim=dim))
+    
+    if any(is_masked_array):
+        return Array(ma.stack(arrs, axis=dim))
+    
+    return Array(np.stack(arrs, axis=dim))
+
 class Array:
     def __init__(self, data):
         """
@@ -201,26 +221,56 @@ class Array:
     def __rpow__(self, other):
         return Array(other ** self.data)
     
+    def __invert__(self):
+        return Array(~self.data)
+    
+    def any(self, dim=None):
+        return Array(self.data.any(dim=dim)) if self.is_tensor else Array(self.data.any(axis=dim))
+    
+    def all(self, dim=None):
+        return Array(self.data.all(dim=dim)) if self.is_tensor else Array(self.data.all(axis=dim))
+    
+    def nonzero(self):
+        out = self.data.nonzero(as_tuple=True) if self.is_tensor else self.data.nonzero()
+        return tuple(Array(out_i) for out_i in out)
+    
     def astype(self, dtype):
         if self.is_tensor:
+            if isinstance(dtype, str):
+                dtype = getattr(torch, dtype)
             arr = Array(self.data.type(dtype))
             if arr.data.storage().data_ptr() == self.data.storage().data_ptr():
                 arr.writeable = self.writeable # same writeable flag since memory is shared
         else:
-            arr = Array(data.astype(dtype))
+            arr = Array(self.data.astype(dtype))
         
         return arr
     
     def asnumpy(self):
         """
-        Coerces underlying data to be np.ndarray. Underlying data is not copied.
+        Coerces underlying tensor data to be np.ndarray. Underlying data is not copied.
         """
         if self.is_tensor:
-            arr = Array(self.data.numpy())
-            arr.writeable = self.writeable # same writeable flag since memory is shared (.numpy() does not copy data)
+            data = self.data.detach().cpu()
+            arr = Array(data.numpy())
+            if data.storage().data_ptr() == self.data.storage().data_ptr():
+                writeable = self.writeable # same writeable flag since memory is shared (.numpy() does not copy data)
             return arr
         return self
     
+    def asndarray(self):
+        """
+        Turns torch.tensor into regular np.ndarray and 
+        turns ma.MaskedArray into regular np.ndarray by casting to nan-compatible dtype and then
+        filling masked entries with nans.
+        """
+        arr = self.asnumpy()
+        if arr.is_masked_array:
+            data = arr.data
+            data = data.astype(np.result_type(data.dtype, float)) # promote to at least float so we can turn masked elements to nan
+            arr = Array(data.filled(np.nan))
+        return arr
+    
     def astensor(self):
         """
         Coerces underlying data to be torch.Tensor. Underlying data is not copied.
@@ -279,13 +329,22 @@ class Array:
             arr = self
         return arr
     
-    def to(self, device):
-        if self.is_tensor:
-            arr = Array(self.data.to(device))
-            if arr.data.storage().data_ptr() == self.data.storage().data_ptr():
-                arr.writeable = self.writeable # same writeable flag since memory is shared
-        else:
-            arr = self
+    def to(self, device, force=False):
+        if device is None:
+            device = 'cpu'
+        if not isinstance(device, str):
+            device = device.type
+            
+        arr = self.astensor() if force and device != 'cpu' else self
+            
+        if arr.is_tensor:
+            if device == 'mps' and arr.dtype == torch.float64:
+                arr = arr.astype(torch.float32) # mps currently does not support float64
+            new_arr = Array(arr.data.to(device))
+            if new_arr.data.storage().data_ptr() == arr.data.storage().data_ptr():
+                new_arr.writeable = arr.writeable # same writeable flag since memory is shared
+            arr = new_arr
+            
         return arr
     
     def cpu(self):
@@ -312,36 +371,57 @@ class Array:
         if self.is_tensor:
             return Array(self.data.isnan())
         else:
-            if isinstance(self.data, np.floating):
-                na = np.isnan(self.data)
+            if self.data.dtype.kind == 'f':
+                # if subok=True, na will be a masked array with the same mask as self.data,
+                # so that setting na[self.data.mask] = True will modify the mask of self.data.
+                # This caused a bug that was really difficult to diagnose
+                na = np.isnan(self.data, subok=False)
             else:
                 na = np.zeros(self.shape, dtype=bool)
             if self.is_masked_array:
-                if isinstance(na, ma.MaskedArray):
-                    na = na.data
                 na[self.data.mask] = True
             return Array(na)
     
-    def unique(self, return_inverse=False, return_counts=False, dim=None):
+    def unique(self, sorted=True, return_index=False, return_inverse=False, return_counts=False, dim=None, equal_nan=False):
         if self.is_tensor:
-            out = self.data.unique(return_inverse=return_inverse, return_counts=return_counts, dim=dim)
+            out = lib.pt.unique(self.data, sorted=sorted, return_index=return_index, return_inverse=return_inverse, return_counts=return_counts, dim=dim, equal_nan=equal_nan)
         elif self.is_masked_array:
-            out = ma.unique(self.data, return_inverse=return_inverse, return_counts=return_counts, axis=dim)
+            if not equal_nan:
+                raise ValueError("equal_nan=True not supported for ma.MaskedArray")
+            if return_counts:
+                raise ValueError("return_counts=True not supported for ma.MaskedArray")
+            if dim is not None:
+                raise ValueError(f"only dim=None is supported for ma.MaskedArray, but {dim=}.")
+            if not sorted:
+                raise ValueError("sorted=False not supported for ma.MaskedArray.")
+            out = ma.unique(self.data, return_index=return_index, return_inverse=return_inverse)
         else:
-            out = np.unique(self.data, return_inverse=return_inverse, return_counts=return_counts, axis=dim)
+            if not sorted:
+                raise ValueError("sorted=False not supported for np.ndarray.")
+            out = np.unique(self.data, return_index=return_index, return_inverse=return_inverse, return_counts=return_counts, axis=dim, equal_nan=equal_nan)
             
         if isinstance(out, tuple):
             return [Array(out_i) for out_i in out]
         return Array(out)
     
-    @unpack_args
-    def bincount(self, weights=None):
+    # @unpack_args
+    def bincount(self, weights=None, minlength=0, nan_policy='propagate'):
+        weights = weights.data if isinstance(weights, Array) else weights
         if self.is_tensor:
-            # arr = Array(self.data.bincount(weights=weights))
+            # arr = Array(self.data.bincount(weights=weights.data))
             ### torch.bincount derivative is not implemented as of torch==1.13.1, so use my custom bincount instead. ###
-            arr = Array(lib.pt.bincount(self.data, weights=weights))
+            arr = Array(lib.pt.bincount(self.data, weights=weights, minlength=minlength, nan_policy=nan_policy))
         else:
-            arr = Array(np.bincount(self.data, weights=weights))
+            x = self.data
+            if self.is_masked_array:
+                mask = x.mask
+                x = x[~mask].data
+                weights = weights[~mask]
+                
+            if isinstance(weights, ma.MaskedArray):
+                weights = Array(weights).asndarray().data
+
+            arr = Array(lib.np.bincount(x, weights=weights, minlength=minlength, nan_policy=nan_policy))
             
         return arr
     
@@ -352,4 +432,12 @@ class Array:
             arr = Array(np.argsort(self.data, axis=dim))
             if descending:
                 arr = arr[::-1]
-            return arr
\ No newline at end of file
+            return arr
+        
+    def bin(self, **kwargs):
+        if self.is_tensor:
+            out = lib.pt.stats.bin(self.data, nan_policy='omit', **kwargs)
+        else:
+            out = lib.sp.stats.bin(self.asndarray().data, nan_policy='omit', **kwargs)
+            
+        return [Array(out_i) for out_i in out]
\ No newline at end of file
