@@ -14,7 +14,7 @@ from . import parsing
 
 logger = logging.getLogger(__name__)
 
-__all__ = ['NA', 'InternalError', 'DataFrame', 'DataFrameGroupBy', 'concat']
+__all__ = ['NA', 'InternalError', 'DataFrame', 'DataFrameGroupBy', 'ColumnGroupBy', 'concat']
 
 NA = ma.masked
 
@@ -190,9 +190,9 @@ class DataFrame:
         for k, v in self._data.items():
             yield k, v.data
         
-    def iterrows(self):
+    def __iter__(self):
         for i in range(len(self)):
-            yield i, self[:,i]
+            yield self[:,i]
      
     @property
     def dtype(self):
@@ -326,12 +326,17 @@ class DataFrame:
     def copy(self):
         # Should work similarly to pd.DataFrame.copy(deep=True)
         return DataFrame(self._data, copy=True)
+    
+    def torch(self):
+        df = DataFrame(self, copy=False)
+        for k, v in df._items():
+            df[k] = v.astensor()
+        return df
         
     def to(self, device):
-        df = self.copy()
-        for k, v in df._items():
-            df[k] = v.to(device)
-        return df
+        for k, v in self._items():
+            self[k] = v.to(device)
+        return self
     
     def to_list(self):
         """
@@ -357,7 +362,7 @@ class DataFrame:
         """
         return np.stack([v.detach().cpu().numpy() for v in self._values()], axis=1) # np.stack returns copy
     
-    def to_torch(self, device=None):
+    def to_tensor(self, device=None):
         """
         Returns data as a 2D torch.Tensor, but may raise error if there is a column with a datatype
         that is not supported by pytorch. Data is copied (but non-array objects are not copied).
@@ -443,6 +448,25 @@ class DataFrame:
         arr = np.stack(arrs, axis=-1)
         return arr
     
+    def _to_numeric(self):
+        info = self.info()
+        is_tensor = any(info['is_tensor'])
+        gpu_device_types = {device.type for device in info['device'] if device.type != 'cpu'}
+        if len(gpu_device_types) > 1:
+            raise ValueError(f"_to_numeric expects only one type of gpu device, but found {gpu_device_types=}.")
+        device = gpu_device_types.pop() if len(gpu_device_types) == 1 else 'cpu'
+        
+        arrs = []
+        for arr in self._values():
+            if not arr.is_tensor and arr.dtype.kind not in 'biufc': # non-numeric
+                _, arr = arr.unique(return_inverse=True, equal_nan=True)
+            if is_tensor:
+                arr = arr.astensor()
+            arr = arr.to(device)
+            arrs.append(arr)
+        arr = ar.stack(arrs, dim=-1)
+        return arr
+    
     def info(self):
         return DataFrame([{
             'column': k,
@@ -471,21 +495,22 @@ class DataFrame:
         # Should works similarly to pd.DataFrame.groupby(by, as_index=False)
         return DataFrameGroupBy(self, *args, **kwargs)
     
-    def where(self, condition, level=0):
+    def eval(self, condition, level=0):
+        condition = parsing.preparse(condition)
         var_names = parsing.parse_var_names(condition)
-        condition = parsing.modify_expr(condition, var_names)
         frame = sys._getframe(level+1) # see https://github.com/pandas-dev/pandas/blob/main/pandas/core/computation/scope.py
+        env = frame.f_globals | frame.f_locals
+        
         # add @ variables
-        var_dict = {f'__eval_local_{var_name}': frame.f_locals[var_name] for var_name in var_names}
+        var_dict = {f'{parsing.LOCAL_TAG}{k}': env[k] for k in var_names}
+        
         # add dataframe columns
-        var_dict = var_dict | {k: v.data for k, v in self._data.items()}
-        # add ability to recognize backtick quoted columns
-        var_dict = var_dict | {f"BACKTICK_QUOTED_STRING_{k.replace(' ', '_')}": v.data for k, v in self._data.items()}
-        result = eval(condition, var_dict)
-        return result
+        var_dict = var_dict | {parsing.clean_column_name(k): v.data for k, v in self._data.items() if not isinstance(k, int)}
+        
+        return eval(condition, var_dict)
     
     def query(self, condition, level=0):
-        return self[:,self.where(condition, level=level+1)]
+        return self[:,self.eval(condition, level=level+1)]
     
     def merge(self, df, how='inner', on=None, left_on=None, right_on=None, suffixes=('_x', '_y')):
         """
@@ -520,36 +545,23 @@ class DataFrame:
         
         larr, linv, lcount = lib.np.unique_rows(larr, sorted=False, return_inverse=True, return_counts=True)
         rarr, rinv, rcount = lib.np.unique_rows(rarr, sorted=False, return_inverse=True, return_counts=True)
-        # logger.debug(f'{larr=}')
-        # logger.debug(f'{linv=}')
-        # logger.debug(f'{lcount=}')
-        # logger.debug(f'{rarr=}')
-        # logger.debug(f'{rinv=}')
-        # logger.debug(f'{rcount=}')
+
         _, larr_idx, rarr_idx = lib.np.intersect_rows(larr, rarr, assume_unique=True, return_indices=True)
-        # logger.debug(f'{larr_idx=}')
-        # logger.debug(f'{rarr_idx=}')
+
         lmask, rmask = np.isin(linv, larr_idx), np.isin(rinv, rarr_idx)
-        # logger.debug(f'{lmask=}')
-        # logger.debug(f'{rmask=}')
+
         lindices, rindices = lmask.nonzero()[0], rmask.nonzero()[0]
-        # logger.debug(f'{lindices=}')
-        # logger.debug(f'{rindices=}')
+
         inv_larr_idx, inv_rarr_idx = np.empty(len(larr), dtype=int), np.empty(len(rarr), dtype=int)
         inv_larr_idx[larr_idx] = np.arange(len(larr_idx))
         inv_rarr_idx[rarr_idx] = np.arange(len(rarr_idx))
-        # logger.debug(f'{inv_larr_idx=}')
-        # logger.debug(f'{inv_rarr_idx=}')
+
         lindices, rindices = lindices[np.argsort(inv_larr_idx[linv[lmask]])], rindices[np.argsort(inv_rarr_idx[rinv[rmask]])]
-        # logger.debug(f'{lindices=}')
-        # logger.debug(f'{rindices=}')
+
         lcount, rcount = lcount[larr_idx], rcount[rarr_idx]
-        # logger.debug(f'{lcount=}')
-        # logger.debug(f'{rcount=}')
+
         lindices = np.repeat(lindices, np.repeat(rcount, lcount))
         rindices = lib.np.repeat(rindices, lcount, chunks=rcount)
-        # logger.debug(f'{lindices=}')
-        # logger.debug(f'{rindices=}')
         
         on = [column for column in left_on if column in right_on]
         left_not_on = [column for column in self.columns if column not in on]
@@ -572,68 +584,107 @@ class DataFrame:
 
         return concat([left.rename(left_rename), middle, right.rename(right_rename)], axis=1)
     
+    def bin(self, column, bins=10, range=None, return_edges=False, return_binnums=False):
+        binnums, centers, edges = self._data[column].bin(bins=bins, range=range)
+        out = [centers[binnums]]
+        
+        if return_edges:
+            ledges = ar.concat([ar.Array([-np.inf]).like(edges), edges])[binnums]
+            redges = ar.concat([edges, ar.Array([ np.inf]).like(edges)])[binnums]
+            out += [ledges, redges]
+            
+        if return_binnums:
+            out += [binnums]
+            
+        if len(out) == 1:
+            return out[0]
+        return tuple(out)
+    
 class DataFrameGroupBy:
-    def __init__(self, df, by, sort=True, dropna=True):
-        assert isinstance(df, DataFrame)
+    def __init__(self, df, by, sort=True, dropna=True, nanstats=True, device=None):
+        # assert isinstance(df, DataFrame)
         self.df = df
         self.by = np.atleast_1d(by)
         self.dropna = dropna
+        self.nanstats = nanstats
         data = self.df[self.by]
         
         if len(data.columns) == 0:
             raise ValueError("No selected columns.")
             
-        arr = self.df[self.by]._to_numpy_numeric() # numeric numpy array representation of selected columns
-        _, idx, inv_idx = lib.np.unique_rows(arr, sorted=sort, return_index=True, return_inverse=True) # faster than np.unique
-        # _, idx, inv_idx = np.unique(arr, return_index=True, return_inverse=True, axis=0)
-        self._groups = DataFrame({column: self.df[column][idx] for column in self.by})
-        self._na = np.stack([arr.isna().detach().cpu().numpy() for arr in self._groups._values()], axis=1).any(axis=1)
-        self._row_to_group_idx = ar.Array(inv_idx)
+        # arr = data._to_numpy_numeric() # numeric numpy array representation of selected columns
+        # _, idx, inv_idx = lib.np.unique_rows(arr, sorted=sort, return_index=True, return_inverse=True) # faster than np.unique
+        # # _, idx, inv_idx = np.unique(arr, return_index=True, return_inverse=True, axis=0) # faster than np.unique
+        # self._groups = DataFrame({column: self.df[column][idx] for column in self.by})
+        # # self._na = np.stack([arr.isna().detach().cpu().numpy() for arr in self._groups._values()], axis=1).any(axis=1)
+        # # self._row_to_group_idx = ar.Array(inv_idx)
+        # self._na = DataFrame({k: v.isna() for k, v in self._groups._items()})._to_numeric().any(dim=1)
+        # self._row_to_group_idx = ar.Array(inv_idx).like(self._na, dtype=False)
+        
+        arr = data._to_numeric().to(device, force=True) # numeric array representation of selected columns
+        if arr.is_masked_array:
+            arr = arr.asndarray() # ma.MaskedArray currently does not support unique
+        _, idx, inv_idx = arr.unique(sorted=sort, return_index=True, return_inverse=True, dim=0)
+        self._groups = DataFrame({column: self.df._data[column][idx] for column in self.by})
+        # self._na = np.stack([arr.isna().asndarray() for arr in self._groups._values()], axis=1).any(axis=1)
+        # self._row_to_group_idx = inv_idx.asndarray()
+        self._na = DataFrame({k: v.isna() for k, v in self._groups._items()})._to_numeric().any(dim=1)
+        self._row_to_group_idx = inv_idx
+        
+        self._group_indexer = GroupIndexer(self)
         
     @property
     def groups(self):
         if self.dropna:
-            return self._groups[:,~self._na]
+            return self._groups[:,~self._na.data]
         return self._groups
     
+    @property
+    def group(self):
+        return self._group_indexer
+    
     def items(self):
-        columns = [column for column in self.df.columns if column not in self.by]
-        for idx, group in self._groups.iterrows():
-            if self.dropna and self._na[idx]:
-                continue
-                
-            yield tuple(group), self.df[columns, self._row_to_group_idx.data == idx]
+        for i, group in enumerate(self.groups):
+            yield group, self.group[i]
             
     def __getitem__(self, key):
         return ColumnGroupBy(self, key)
         
     def agg(self, *args, **kwargs):
         new_args = collections.defaultdict(list)
-        for arg in args:
+        for i, arg in enumerate(args):
             if not (isinstance(arg, tuple) and len(arg) == 2):
                 raise TypeError(f"arg must be a 2-tuple of (column_name, func), but {arg=}.")
                 
             column, func = arg
-            new_args[column].append(func)
+            new_args[column].append((i, func))
         
-        new_kwargs = collections.defaultdict(dict)
-        for k, v in kwargs.items():
+        new_kwargs = collections.defaultdict(list)
+        for i, (k, v) in enumerate(kwargs.items()):
             if not (isinstance(v, tuple) and len(v) == 2):
                 raise TypeError(f"v must be a 2-tuple of (column_name, func), but {v=}.")
                 
             column, func = v
-            new_kwargs[column][k] = func
+            new_kwargs[column].append((i + len(args), (k, func)))
         
         results = {}
+        orders = {}
         for column, args, kwargs in lib.itertools.dict_zip(new_args, new_kwargs, mode='union'):
-            args = [] if args is None else args
-            kwargs = {} if kwargs is None else kwargs
-            result = self[column].agg(*args, **kwargs)
-            if isinstance(result, DataFrame):
-                results = results | {k if k in kwargs else f'{column}_{k}': v for k, v in result.items()}
-            else:
-                results[column] = result
+            arg_order, args = ([], []) if args is None else tuple(zip(*args))
+            kwarg_order, kwargs = ([], []) if kwargs is None else tuple(zip(*kwargs))
+            order = list(arg_order) + list(kwarg_order)
             
+            result = self[column].agg(*args, **dict(kwargs))
+            if isinstance(result, DataFrame):
+                result = {k if k in kwargs else f'{column}_{k}': v for k, v in result.items()}
+            else:
+                result = {column: result}
+            order = {k: v for k, v in zip(result.keys(), order)}
+                
+            results |= result
+            orders |= order
+            
+        results = {k: results[k] for k, _ in sorted(orders.items(), key=lambda item: item[1])}
         return DataFrame(self.groups.as_dict() | results)
     
 class ColumnGroupBy:
@@ -643,21 +694,23 @@ class ColumnGroupBy:
         self._na = dfgb._na
         self.groups = dfgb.groups
         self.dropna = dfgb.dropna
+        self.nan_policy = 'omit' if dfgb.nanstats else 'propagate'
         self.df = dfgb.df
         self.column = column
+        self.data = dfgb.df[column]
         
     def _agg(self, func, results):
         if func in results:
             return results[func]
-        
-        _row_to_group_idx = self._row_to_group_idx.like(self.df[self.column], dtype=False)
 
+        _row_to_group_idx = self._row_to_group_idx.like(self.data, dtype=False)
+        
         # See scipy.stats.binned_statistic_dd for reference
         if func == 'count':
-            result = _row_to_group_idx.bincount()
+            result = _row_to_group_idx.bincount(weights=(~ar.Array(self.data).isna()).astype('int64').data, nan_policy=self.nan_policy)
         
         elif func == 'sum':
-            result = _row_to_group_idx.bincount(weights=self.df[self.column])
+            result = _row_to_group_idx.bincount(weights=self.data, nan_policy=self.nan_policy)
         
         elif func == 'mean':
             result = self._agg('sum', results) / self._agg('count', results)
@@ -665,7 +718,8 @@ class ColumnGroupBy:
         elif func == 'var':
             count, mean = self._agg('count', results), self._agg('mean', results)
             result = (_row_to_group_idx.bincount(
-                weights=(self.df[self.column] - mean[_row_to_group_idx].data)**2
+                weights=(self.data - mean[_row_to_group_idx].data)**2,
+                nan_policy=self.nan_policy
             ))/(count-1)
             
         elif func == 'std':
@@ -677,7 +731,7 @@ class ColumnGroupBy:
             result = std / count**0.5
         
         elif func == 'min':
-            v = ar.Array(self.df[self.column])
+            v = ar.Array(self.data)
             i = v.argsort(descending=True)
             result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
             
@@ -689,7 +743,7 @@ class ColumnGroupBy:
                 torch.use_deterministic_algorithms(False)
         
         elif func == 'max':
-            v = ar.Array(self.df[self.column])
+            v = ar.Array(self.data)
             i = v.argsort()
             result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
             
@@ -701,7 +755,7 @@ class ColumnGroupBy:
                 torch.use_deterministic_algorithms(False)
         
         elif isinstance(func, callable):
-            v = ar.Array(self.df[self.column])
+            v = ar.Array(self.data)
             result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
             for idx, group in enumerate(self._groups):
                 result[idx] = func(v[_row_to_group_idx == idx].data)
@@ -729,6 +783,19 @@ class ColumnGroupBy:
         if len(results) == 1:
             return next(iter(results.values()))
         return DataFrame(results)
+    
+class GroupIndexer:
+    def __init__(self, groupby):
+        self.groupby = groupby
+        self.columns = [column for column in groupby.df.columns if column not in groupby.by]
+        if groupby.dropna:
+            self.idx_map = (~groupby._na).nonzero()[0]
+        
+    def __getitem__(self, idx):
+        if self.groupby.dropna:
+            idx = self.idx_map[idx]
+            
+        return self.groupby.df[self.columns, self.groupby._row_to_group_idx.data == idx]
         
 def concat(dfs, axis=0):
     assert len(dfs) > 0
